import fs from 'fs-extra'
import path from 'path'
import ora from 'ora'
import url from 'url'
import kebabCase from 'lodash.kebabcase'
import { print } from '@toptal/davinci-cli-shared'
import { createRequire } from 'module'

import workspace from '../utils/workspace.js'
// eslint-disable-next-line no-restricted-syntax
import skeleton from '../utils/skeleton.js'
import { LIBS_PATH, NAMESPACES_PATH } from '../constants.js'

const localRequire = createRequire(import.meta.url)
const thisScriptDirname = path.dirname(url.fileURLToPath(import.meta.url))

const getLibPath = (libName, namespace) =>
  namespace
    ? path.join(NAMESPACES_PATH, namespace, LIBS_PATH, libName)
    : path.join(LIBS_PATH, libName)

const checkPrerequisites = async (monorepoWorkspaceDir, libName, namespace) => {
  const spinner = ora('Checking prerequisites...').start()

  // check if run in monorepo
  const isWorkspace = Boolean(monorepoWorkspaceDir)

  if (!isWorkspace) {
    spinner.fail()
    throw new Error(
      "It looks like you run 'new-lib' Davinci command not in a monorepo."
    )
  }

  // check if app already exists
  const libPath = path.join(
    monorepoWorkspaceDir,
    getLibPath(libName, namespace)
  )
  const doesLibExist = await fs.exists(libPath)

  if (doesLibExist) {
    spinner.fail()
    throw new Error(`Library already exists in the folder ${libPath}!`)
  }

  spinner.succeed('Prerequisites checked!')
}

const copyLibrarySkeleton = async (destDir, currentDavinciRunningDir) => {
  const skeletonLibPath = localRequire.resolve('@toptal/davinci-skeleton-lib', {
    paths: [currentDavinciRunningDir],
  })
  const appDir = path.join(skeletonLibPath, '../..')

  const spinner = ora('Copying skeleton...').start()

  try {
    await skeleton.copyAllFiles(appDir, destDir)
    await skeleton.copyTemplatePackageJsonFile(destDir)

    spinner.succeed('Skeleton copied!')
  } catch (err) {
    spinner.fail()
    throw err
  }
}

const modifyPackageJson = async (
  destDir,
  libName,
  { namespace, workspacePackageName }
) => {
  const spinner = ora('Modifying package.json...').start()

  try {
    const config = skeleton.readPackageJson(destDir)

    config.name = namespace
      ? `@toptal/${namespace}-${libName}`
      : `@toptal/${libName}`
    config.homepage = `https://github.com/toptal/${workspacePackageName}/tree/master/${getLibPath(
      libName,
      namespace
    )}`

    delete config.gitHead

    delete config.publishConfig
    config.license = 'SEE LICENSE IN LICENSE.MD'

    config.version = '0.0.1'

    skeleton.writePackageJson(destDir, config)

    spinner.succeed('package.json modified!')
  } catch (err) {
    spinner.fail()
    throw err
  }
}

const installDependencies = async monorepoWorkspaceDir => {
  const spinner = ora('Installing dependencies... ').start()

  try {
    await workspace.installDependencies(monorepoWorkspaceDir)

    spinner.succeed('Dependencies installed!')
  } catch (err) {
    spinner.fail()
    throw new Error(err)
  }
}

const createCommand = async (libName, namespace) => {
  print.header(
    'Creating',
    libName,
    namespace ? `in namespace ${namespace}` : '',
    '...'
  )

  const monorepoWorkspaceDir = workspace.getRootDir()
  const workspacePackageJson = workspace.readPackageJson(monorepoWorkspaceDir)
  const currentDavinciRunningDir = thisScriptDirname

  try {
    // check if all apps are installed, directory doesn't exists, etc.
    await checkPrerequisites(monorepoWorkspaceDir, libName, namespace)

    const destDir = path.join(
      monorepoWorkspaceDir,
      getLibPath(libName, namespace)
    )
    const workspacePackageName = workspace.getPackageName(workspacePackageJson)

    // copy skeleton to destination directory
    await copyLibrarySkeleton(destDir, currentDavinciRunningDir)

    // change package.json
    await modifyPackageJson(destDir, libName, {
      namespace,
      workspacePackageName,
    })

    // run yarn
    await installDependencies(monorepoWorkspaceDir)

    print.success('Library has been created.')
  } catch (err) {
    print.red(err)
    process.exit(1)
  }
}

const newLibCommandCreator = {
  command: 'new-lib',
  description: 'Create new library in monorepo',
  args: [['<libName>', 'Library Name: folder name for the new library']],
  options: [
    {
      label:
        "Namespace of the lib. Creates the lib in a monorepo namespace: './namespaces/{namespace}/libs/{libName}'",
      name: '--namespace <namespace>',
    },
  ],
  action: async (libName, { namespace }) => {
    const formattedLibName = kebabCase(libName)

    return createCommand(formattedLibName, namespace)
  },
}

export default newLibCommandCreator
